import dataclasses
import typing

from unitelabs.cdk import sila


class InvalidCommandSequence(Exception):
    """
    The issued command does not follow the sequence of commands for the device according to its role in the labware
    transfer.
    """


class LabwareNotPicked(Exception):
    """Picking up the labware item from the source device failed."""


class LabwareNotPlaced(Exception):
    """Placing the labware item at the destination device failed."""


@dataclasses.dataclass
class PositionIndex(sila.CustomDataType):
    """
    Specifies a position via an index number, starting at 1.

    .. parameter:: Position index number.
    """

    PositionIndex: typing.Annotated[int, sila.constraints.MinimalInclusive(value=1)]


@dataclasses.dataclass
class HandoverPosition(sila.CustomDataType):
    """
    Specifies one of the possible positions of a device where labware items can be handed over. Can contain a
    sub-position, e.g. for specifying a position in a rack.

    .. parameter:: The name of the handover position (must be unique within the device).
    .. parameter:: The index of a sub-position within a handover position or the number of sub-positions respectively,
                   e.g. for a rack.
    """

    Position: str

    PositionIndex: PositionIndex



import abc
import logging
import typing
from dataclasses import dataclass
from typing import Iterable, Optional

from unitelabs.cdk import sila

logger = logging.getLogger(__name__)


class TaskError(Exception):
    """Error during workflow"""


@dataclass
class TaskStatus(sila.CustomDataType):
    """Task status dataclass

    .. parameter:: Name of the workflow
    .. parameter:: Task identifier
    .. parameter:: Execution status
    .. parameter:: Start time (unix timestamp, utc)
    """

    name: str
    identifier_0: str
    status: str
    start_time: float


# TODO: Refactor this function to prevent C901
# flake8: noqa: C901
def get_workflow_feature(allowed_workflow_names: Optional[Iterable[str]] = None):
    """
    Function to create a WorkflowRunnerService feature

    .. allowed_workflow_names:: List of allowed workflow names. If None, any name is allowed.
    """
    if allowed_workflow_names:
        # Only allow the specified workflow names
        wf_name_annotation = sila.constraints.Set(list(allowed_workflow_names))
    else:
        # Allow any workflow name (len(name) > 0)
        wf_name_annotation = sila.constraints.MinimalLength(1)

    class WorkflowRunnerService(sila.Feature, metaclass=abc.ABCMeta):
        def __init__(
            self,
            *args,
            identifier: str = "WorkflowRunnerService",
            display_name: str = "Workflow Runner Service",
            description: str = "Feature for starting/monitoring long running tasks",
            **kwargs,
        ):
            super().__init__(
                *args,
                identifier=identifier,
                display_name=display_name,
                description=description,
                **kwargs,
            )

        @abc.abstractmethod
        @sila.UnobservableCommand(
            name="Cancel Task",
            errors=[TaskError],
        )
        async def cancel_task(self, identifier: str) -> None:
            """
            Cancel the task with the given identifier

            .. parameter:: Task identifier
            """
            pass

        @abc.abstractmethod
        @sila.UnobservableProperty(
            name="Running Tasks",
        )
        async def get_running_tasks(self) -> list[TaskStatus]:
            """ "Get the status of all running tasks"""
            pass

        @abc.abstractmethod
        @sila.UnobservableCommand(
            name="Task Status",
            errors=[TaskError],
        )
        @sila.Response("Status", "Status of a task")
        async def get_task_status(self, identifier: str) -> str:
            """
            Get the status of the running task

            .. identifier:: Task identifier
            """
            pass

        @abc.abstractmethod
        @sila.UnobservableCommand(
            name="Start New Task",
            errors=[TaskError],
        )
        @sila.Response("Identifier", "Identifier of the started workflow process")
        async def start_new_task(
            self,
            name: typing.Annotated[str, wf_name_annotation],
            arguments_json: str,
        ) -> str:
            """
            "Start a task from a named workflow and return immediately"

            .. name:: Name of the workflow
            .. arguments_json:: JSON encoded arguments for the workflow
            """
            pass

    return WorkflowRunnerService



import abc
import logging
import typing

from unitelabs.cdk import sila

from .types.sila_types import (
    HandoverPosition,
    InvalidCommandSequence,
    LabwareNotPicked,
    LabwareNotPlaced,
    PositionIndex,
)

logger = logging.getLogger(__name__)


class LabwareTransferManipulatorControllerBase(sila.Feature, metaclass=abc.ABCMeta):
    """
    This feature (together with the "Labware Transfer Site Controller" feature) provides commands to trigger the
    sub-tasks of handing over a labware item, e.g. a microtiter plate or a tube, from one device to another in a
    standardized and generic way.

    For each labware transfer a defined sequence of commands has to be called on both involved devices to ensure the
    proper synchronization of all necessary transfer actions without unwanted physical interferences and to optimize
    the transfer performance regarding the execution time. Using the generic commands, labware transfers between any
    arbitrary labware handling devices can be controlled (a robot device has not necessarily to be involved).

    Generally, a labware transfer is executed between a source and a destination device, where one of them is the
    active device (executing the handover actions) and the other one is the passive device.

    The "Labware Transfer Manipulator Controller" feature is used to control the labware transfer on the side of the
    active device to hand over labware to or take over labware from a passive device, which provides the
    "Labware Transfer Site Controller" feature.

    If a device is capable to act either as the active or as the passive device of a labware transfer it must provide
    both features.

    The complete sequence of issued transfer commands on both devices is as follows:

    1. Prior to the actual labware transfer a "Prepare For Output" command is sent to the source device to execute all
       necessary actions to be ready to release a labware item (e.g. open a tray) and simultaneously a "Prepare For
       Input" command is sent to the destination device to execute all necessary actions to be ready to receive a
       labware item (e.g. position the robotic arm near the tray of the source device).
    2. When both devices have successfully finished their "Prepare For ..." command execution, the next commands are
       issued.
    3a If the source device is the active device it will receive a "Put Labware" command to execute all necessary
       actions to put the labware item into the destination device. After the transfer has been finished successfully,
       the destination device receives a "Labware Delivered" command, that triggers all actions to be done after the
       labware item has been transferred (e.g. close the opened tray).
    3b If the destination device is the active device it will receive a "Get Labware" command to execute all necessary
       actions to get the labware from the source device (e.g. gripping the labware item). After that command has been
       finished successfully, the source device receives a "Labware Removed" command, that triggers all actions to be
       done after the labware item has been transferred (e.g. close the opened tray).

    The command sequences for an active source or destination device have always to be as follows:
    - for an active source device:        PrepareForOutput - PutLabware.
    - for an active destination device:   PrepareForInput - GetLabware.

    If the commands issued by the client differ from the respective command sequences an "Invalid Command Sequence"
    error will be raised.

    To address the location, where a labware item can be handed over to or from other devices, every device must
    provide one or more uniquely named positions (handover positions) via the "Available Handover Positions" property.
    A robot (active device) should have at least one handover position for each device that it interacts with, whereas
    most passive devices will only have one handover position. In the case of a position array (e.g. a rack), the
    position within the array is specified via the sub-position of the handover position, passed as an index number.

    To address the positions within a device where the transferred labware item has to be stored at or is to be taken
    from (e.g. the storage positions inside an incubator), the internal position is specified. Each device must provide
    the number of available internal positions via the "Number Of Internal Positions" property. In the case of no
    multiple internal positions, this property as well as the "Internal Position" parameter value must be 1.

    With the "Prepare For Input" command there is also information about the labware transferred, like labware type or
    a unique labware identifier (e.g. a barcode).

    The "Intermediate Actions" parameter of the "Put Labware" and "Get Labware" commands can be used to specify commands
    that have to be executed while a labware item is transferred to avoid unnecessary movements, e.g. if a robot has to
    get a plate from a just opened tray and a lid has to be put on the plate before it will be gripped, the lid handling
    actions have to be included in the "Get Labware" actions. The intermediate actions have to be executed in the same
    order they have been specified by the "Intermediate Actions" parameter.
    The property "Available Intermediate Actions" returns a list of commands that can be included in a "Put Labware" or
    "Get Labware" command.
    """

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="instruments.labware.manipulation",
            version="1.0",
            maturity_level="Verified",
        )

    @abc.abstractmethod
    @sila.ObservableCommand(name="Prepare For Input", errors=[InvalidCommandSequence])
    async def PrepareForInput(
        self,
        HandoverPosition: HandoverPosition,
        InternalPosition: PositionIndex,
        LabwareType: str,
        LabwareUniqueID: str,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to accept new labware at the specified handover position.

        .. parameter:: Indicates the position where the labware will be handed over.
        .. parameter:: Indicates the position which the labware will be stored at within the device, e.g. internal
                       storage positions of an incubator.
        .. parameter:: Specifies the type of labware that will be handed over to transfer information about the labware
                       that the device might need to handle it correctly.
        .. parameter:: Represents the unique identification of a labware in the controlling system. It is assigned by
                       the system and must remain unchanged during the whole process.
           :display_name: Labware Unique ID
        """

    @sila.ObservableCommand(name="Prepare For Output", errors=[InvalidCommandSequence])
    async def PrepareForOutput(
        self,
        HandoverPosition: HandoverPosition,
        InternalPosition: PositionIndex,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to release the labware at the specified handover position.

        .. parameter:: Indicates the position where the labware will be handed over.
        .. parameter:: Indicates the position which the labware will be retrieved from within the device, e.g. internal
                       storage positions of an incubator.
        """
        logger.info("Running PrepareForOutput")

    @abc.abstractmethod
    @sila.ObservableCommand(
        name="Put Labware", errors=[InvalidCommandSequence, LabwareNotPlaced]
    )
    async def PutLabware(
        self,
        HandoverPosition: HandoverPosition,
        IntermediateActions: list[str],
        *,
        status: sila.Status,
    ) -> None:
        """
        Place the currently processed labware item at the specified handover position (sent to the active source device
        after a "Prepare For Output" command).

        .. parameter:: Indicates the position the labware item will be moved to.
        .. parameter:: Specifies one or more commands that have to be executed within the command sequence (e.g.
                       removing a lid).
                       The order of execution is specified by order within the given list.
                       Each entry must be one of the commands returned by the AvailableIntermediateCommandExecutions
                       property.
        """

    @abc.abstractmethod
    @sila.ObservableCommand(
        name="Get Labware", errors=[InvalidCommandSequence, LabwareNotPicked]
    )
    @sila.Response(
        "HandoverPosition", "The position where the labware was retrieved from."
    )
    async def GetLabware(
        self,
        HandoverPosition: HandoverPosition,
        IntermediateActions: list[str],
        *,
        status: sila.Status,
    ) -> HandoverPosition:
        """
        Retrieve a labware item from the specified handover position (sent to the active destination device after a
        "Prepare For Input" command).

        .. parameter:: Indicates the position where the labware will be retrieved from.
        .. parameter:: Specifies one or more commands that have to be executed within the command sequence (e.g.
                       removing a lid).
                       The order of execution is specified by order within the given list.
                       Each entry must be one of the commands returned by the AvailableIntermediateCommandExecutions
                       property.
        """

    @abc.abstractmethod
    @sila.UnobservableProperty(name="Available Handover Positions")
    async def AvailableHandoverPositions(self) -> list[HandoverPosition]:
        """All handover positions of the device including the number of sub-positions."""

    @sila.UnobservableProperty(name="Number Of Internal Positions")
    async def NumberOfInternalPositions(
        self,
    ) -> typing.Annotated[int, sila.constraints.MinimalInclusive(value=1)]:
        """The number of addressable internal positions of the device."""
        return 1  # Default (not used)

    @sila.UnobservableProperty(name="Available Intermediate Actions")
    async def AvailableIntermediateActions(
        self,
    ) -> list[
        typing.Annotated[
            str,
            sila.constraints.FullyQualifiedIdentifier(value="CommandIdentifier"),
        ]
    ]:
        """Returns all commands that can be executed within a "Put Labware" or "Get Labware" command execution."""
        return []  # Default (not used)



import asyncio
import time

from grpc import ChannelConnectivity
from grpc.aio import insecure_channel


class GRPC:
    """Helper class/function made for pinging GRPC style services, e.g. SiLA services

    This is way faster than using a full `sila2.client.SilaClient` instance to connect.
    """

    @classmethod
    async def PingSila(cls, url: str, timeout: float = 0.5):
        channel = insecure_channel

        start = time.time()
        async with channel(url) as ch:
            while time.time() - start < timeout:
                await asyncio.sleep(0.2)
                if ch.get_state(True) == ChannelConnectivity.READY:
                    return True
            return False



from __future__ import annotations

from pathlib import Path
from typing import Any, Optional, Union


def dynamic_import_config(config: Optional[Union[str, Path, dict[str, Any]]]) -> list:
    """
    Similar to dynamic_import, but takes the path to a config file or a dict as input. \
    The config file need contain a single dictionary, on call the functions/classes \
        are imported and instantiated. Should config be `None`, an empty list is returned.

    Args:
        config (str | Path | dict | None): Path to a config file or the config content.


    Supported formats:
    - JSON
    - YAML

    Example config dict:
    ```
    config = {
        "item1": {
            "import": "module1:item1",
            "kwargs": {
                "key1": "value1",
            },
            "args": ["arg1", "arg2"]
        }
    }
    ```

    Example code:

    ```python
    # The code:
    for feature in dynamic_import_config("config.json"):
        server.add(feature)

    # is equivalent to:
    from module1 import item1
    server.add(item1("arg1", "arg2", key1="value1"))
    ```
    """

    # If no config is provided, return an empty list
    if config is None:
        return []  # Nothing

    # If the config is a dictionary, assume it is the content
    if isinstance(config, dict):
        # Load and instantiate the content
        return [
            dynamic_import(c["import"])(*c.get("args", []), **c.get("kwargs", {}))
            for c in config.values()
        ]

    # Config is a path
    # Make sure the path is a Path object
    if isinstance(config, str):
        config = Path(config)

    if config.suffix in (".json",):  # Handle JSON files
        import json

        load = json.load
    elif config.suffix in (".yaml", ".yml"):  # Handle YAML files
        try:
            import yaml
        except ImportError:
            raise ImportError(
                "PyYAML is required to load YAML files. Install it with 'pip install pyyaml'"
            )

        load = yaml.safe_load

    else:
        raise ValueError(f"Unsupported config file format '{config}'")

    # Load the config file
    with open(config, "r") as f:
        config = load(f)

    # Type check
    if not isinstance(config, dict):
        raise ValueError("Config file must contain a dictionary")

    # Load and instantiate the content
    return [
        dynamic_import(c["import"])(*c.get("args", []), **c.get("kwargs", {}))
        for c in config.values()
    ]


def dynamic_import(name: str):
    """Dynamically load an item from a python module.

    ```python
    # Import example:
    my_feature = dynamic_import("my_module:my_feature")
    server.add(my_feature()) # Instantiate and add the feature to the server

    # Content of my_module.py:
    class my_feature(sila.Feature):
        pass
    ```


    Args:
        name (str): The name of the item to load in format 'module:item'.

    """
    from importlib import import_module

    try:
        module, item = name.split(":")
    except ValueError:
        raise ValueError(f"Invalid name '{name}'. Need to be in format 'module:item'")

    try:
        return getattr(import_module(module), item)
    except AttributeError as ex:
        raise AttributeError(f"Item '{item}' not found in module '{module}'") from ex


def get_xml(name: str, version: str) -> str:
    return get_xml_path(name, version).read_text()


def get_xml_path(name: str, version: str) -> Path:
    p = Path(__file__) / ".." / Path(name) / Path(version) / f"{name}.sila.xml"
    p = p.resolve(strict=False)  # Clean-up path and validate format

    if not Path(*p.parts[:-2]).exists():
        raise FileNotFoundError(f"Feature name '{name}' not found")
    if not Path(*p.parts[:-1]).exists():
        raise FileNotFoundError(f"Feature version '{version}' not found")
    if not p.exists():
        ex = FileNotFoundError(f"Xml missing at path '{p}'")
        raise ex

    return p



from unittest import TestCase

from sila2_feature_lib import dynamic_import, dynamic_import_config


class TestDynamicImport(TestCase):
    def test_dynamic_import(self):
        """Test dynamic import"""
        f = dynamic_import("resources.import_me:func1")
        self.assertEqual(f.__name__, "func1")

    def test_dynamic_import_config(self):
        """Test dynamic import config"""
        items = dynamic_import_config("tests/resources/import_config.json")
        self.assertEqual(len(items), 1)  # On item in the config file
        self.assertEqual(type(items[0]).__name__, "MyClass")  # Name match
        self.assertEqual(items[0].args, (1, 2, 3))
        self.assertEqual(items[0].kwargs, {"a": 1, "b": 2})



# This feature tracks external resources (REST APIs, other SiLA services) and their status

import asyncio
import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Type, Union

from unitelabs.cdk import sila

from .util.communication import BaseCall, RestCall, SilaCall

logger = logging.getLogger(__name__)


@dataclass
class Resource(sila.CustomDataType):
    """Resource definition

    .. parameter:: Resource name/alias
    .. parameter:: Url
    .. parameter:: Type
    .. parameter:: Simulated
    """

    alias: str
    url: str
    type: str
    sim: bool = False

    async def Call(self, method: str, *args, **kwargs) -> Any:
        return await GlobalResources.GetResourceType(self.type)(
            alias=self.alias,
            url=self.url,
            method=method,
            simulated=self.sim,
        ).Call(*args, **kwargs)

    async def Check(self) -> bool:
        return await GlobalResources.GetResourceType(self.type)(
            alias=self.alias,
            url=self.url,
            method="test",
            simulated=self.sim,
        ).Ping()


@dataclass
class ResourceStatus(sila.CustomDataType):
    """Resource status

    .. parameter:: Resource name/alias
    .. parameter:: Status
    """

    alias: str
    status: str


class GlobalResources:
    """
    Singleton class to manage/access global resources defined through the ResourcesService Feature
    """

    # Static field of references to the available resources
    __resources: dict[str, Resource] = {}

    # Static field of references to the available resource types
    __type_registry: dict[str, Type[BaseCall]] = {
        "restapi": RestCall,  # Built-in resource type -> REST API
        "sila2": SilaCall,  # Built-in resource type -> SiLA2 Server
    }

    @classmethod
    def Add(cls, resource: Resource):
        cls.__resources[resource.alias] = resource

    @classmethod
    def Get(cls) -> List[Resource]:
        return list(cls.__resources.values())

    @classmethod
    def GetResourceType(cls, name: str) -> Type[BaseCall]:
        return cls.__type_registry[name]

    @classmethod
    def GetItem(cls, filter: str):
        return cls.__resources[filter]

    @classmethod
    def GetHandle(cls, alias: str):
        """Returns a function mapped on the named resource"""

        async def get(method: str, *args, **kwargs) -> Any:
            try:
                resource = cls.GetItem(alias)
            except KeyError:
                logger.error(f"Resource '{alias}' not found")
                raise
            except Exception:
                logger.error(f"Error getting resource {alias}")
                raise
            try:
                return await resource.Call(method, *args, **kwargs)
            except Exception as e:
                logger.error(f"Error calling resource {alias} with method {method}")
                logger.exception(e)
                raise

        return get

    @classmethod
    def RegisterResourceType(cls, name: str, new_type: Type[BaseCall]):
        if name in cls.__type_registry:
            raise ValueError(f"Resource type '{name}' already registered")
        cls.__type_registry.update({name: new_type})

    @classmethod
    async def Status(cls) -> List[ResourceStatus]:
        resources = cls.Get()
        tasks = [r.Check() for r in resources]

        return [
            ResourceStatus(alias=r.alias, status="OK" if t else "ERROR")
            for t, r in zip(await asyncio.gather(*tasks), resources)
        ]


class ResourcesService(sila.Feature):
    def __init__(self, config: Union[Path, str]):
        super().__init__(description="Resource service")
        self.GR = GlobalResources  # Static reference
        self._load_resources(config)

    def _load_resources(self, path: Union[Path, str]):
        if not Path(path).exists():
            raise FileNotFoundError(f"Config file not found: {path}")
        if not Path(path).is_file():
            raise FileNotFoundError(f"Config file is not a file: {path}")

        def yaml_import_load(f):
            import yaml

            return yaml.safe_load(f)

        with open(Path(path), "r") as f:
            config: dict = {
                ".yaml": yaml_import_load,
                ".yml": yaml_import_load,
                ".json": json.load,
            }[Path(path).suffix](f)

        root: dict[str, dict] = config["resources"]

        for item in root.values():
            self.GR.Add(Resource(**item))

    @sila.UnobservableProperty()
    async def Resources(self) -> List[Resource]:
        return self.GR.Get()

    @sila.UnobservableProperty()
    async def Status(self) -> List[ResourceStatus]:
        return await self.GR.Status()



import logging
from os import environ
from typing import Callable

try:
    from unitelabs.cdk import sila
except ImportError as ex:
    raise ImportError(
        "Please install the unitelabs package by running 'pip install sila2-feature-lib[unitelabs]'"
    ) from ex

logger = logging.getLogger(__name__)


class StartSimulationModeFailed(Exception):
    """The server cannot change to Simulation Mode.
    This error can, e.g., be thrown, if a real-world process needs to be ended before switching to simulation \
    mode."""


class StartRealModeFailed(Exception):
    """The server cannot change to Real Mode.
    This error can, e.g. be thrown, if a device is not ready to change into Real Mode.
    """


class SimulationModeGlobal:
    _state = {"active": False, "initialized": False, "cbs": []}

    @classmethod
    def _run_cb(cls, state: bool):
        for cb in cls._state["cbs"]:
            try:
                cb(state)
            except Exception as e:
                logger.exception(f"Error in simulation mode change callback: {e}")

    @classmethod
    def set_simulation_active(cls):
        logger.info("Setting simulation mode")
        if not cls._state["active"]:
            cls._state["active"] = True
            cls._run_cb(True)

    @classmethod
    def set_simulation_inactive(cls):
        logger.info("Setting real mode")
        if cls._state["active"]:
            cls._state["active"] = False
            cls._run_cb(False)

    @classmethod
    def is_simulation_active(cls) -> bool:
        return cls._state["active"]

    @classmethod
    def initialize(cls, state: bool = False):
        if cls._state["initialized"]:
            logger.exception("SimulationModeGlobal already initialized")
            raise RuntimeError("SimulationModeGlobal already initialized")
        cls._state["initialized"] = True
        logger.info(f"Setting initial simulation mode to {state}")
        cls._state["active"] = state
        cls._run_cb(state)

    @classmethod
    def register_on_change(cls, cb: Callable[[bool], None]):
        cls._state["cbs"].append(cb)


# Alias
SMG = SimulationModeGlobal


class SimulatorController(sila.Feature):
    def __init__(self):
        SMG.initialize(
            False if environ.get("APP_SIMULATION_MODE", None) is None else True
        )
        super().__init__(
            identifier="SimulationController",
            display_name="Simulation Controller",
            description="""This Feature provides control over the simulation behaviour of a SiLA Server.

    A SiLA Server can run in two modes:
    (a) Real Mode - with real activities, e.g. addressing or controlling real hardware, e.g. through serial/CANBus commands,
        writing to real databases, moving real objects etc.
    (b) Simulation Mode - where every command is only simulated and responses are just example returns.

    Note that certain commands and properties might not be affected by this feature if they
    do not interact with the real world.""",
        )

    @sila.UnobservableCommand(
        errors=[StartSimulationModeFailed],
    )
    async def start_simulation_mode(self) -> None:
        """Sets the SiLA Server to run in Simulation Mode, i.e. all following commands are executed in simulation
         mode.

        The Simulation Mode can only be entered, if all hardware operations have been safely terminated
        or are in a controlled, safe state.

        The simulation mode can be stopped by issuing the 'Start Real Mode' command."""
        SMG.set_simulation_active()

    @sila.UnobservableCommand(
        errors=[StartRealModeFailed],
    )
    async def start_real_mode(self) -> None:
        """Sets the SiLA Server to run in real mode, i.e. all following commands are executed with real-world
        interactions, like serial port/CAN communication, motor actions etc.

        If the server is in Simulation Mode it can be interrupted at any time. A re-initialization of
        the hardware might be required. The Real Mode can be stopped by issuing the 'Start Simulation Mode' command.
        """
        SMG.set_simulation_inactive()

    @sila.UnobservableProperty()
    async def simulation_mode(self) -> bool:
        """Indication whether SiLA Server is in Simulation Mode or not."""
        return SMG.is_simulation_active()



import asyncio
import logging
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from time import sleep as nonasync_sleep

import httpx
from sila2.client import ClientObservableCommand as COC
from sila2.client import ClientObservableCommandInstance as COCI
from sila2.client import (
    ClientObservableCommandInstanceWithIntermediateResponses as COCIWIR,
)
from sila2.client import ClientObservableProperty as COP
from sila2.client import ClientUnobservableCommand as CUC
from sila2.client import ClientUnobservableProperty as CUP
from sila2.client import SilaClient
from sila2.framework.errors.defined_execution_error import DefinedExecutionError
from sila2.framework.errors.undefined_execution_error import UndefinedExecutionError

from .grpc import GRPC

logger = logging.getLogger(__name__)


async def run_non_blocking(func, *args, **kwargs):
    return await asyncio.get_running_loop().run_in_executor(None, func, *args, **kwargs)


@dataclass
class BaseCall(metaclass=ABCMeta):
    """Base class for communication calls"""

    alias: str
    url: str
    method: str
    simulated: bool = False

    @abstractmethod
    async def Call(self, *args, **kwargs):
        pass

    async def Ping(self) -> bool:
        return True  # Default - test successful


class RestCall(BaseCall):
    """Simple resource implementation for making REST calls"""

    async def Call(self, *args, **kwargs):
        if self.simulated:  # Simulation active!!
            logger.info("Simulating call to: %s", self.method)
            return

        call, *sub = self.method.split(".")
        rpl: httpx.Response = await getattr(httpx, call)(*sub, *args, **kwargs)
        return rpl.raise_for_status().json()

    async def Ping(self) -> bool:
        try:
            async with httpx.AsyncClient(timeout=2) as client:
                rpl = await client.get(f"http://{self.url}/docs", timeout=2)
                rpl.raise_for_status()
                return True
        except Exception:
            return False


@contextmanager
def sila_call_cash():
    """Context manager for cashing SiLA calls, speeds up communication by sharing a single client instance between calls.

    ```python
    with sila_call_cash():
        data = await sila_resource("feature1:GetData1")
        data = await sila_resource("feature1:GetData2")
        data = await sila_resource("feature1:GetData3")
        data = await sila_resource("feature2:GetData1")
    ```
    """

    SilaCall.__context_counter__["counter"] += 1
    try:
        yield
    finally:
        SilaCall.__context_counter__["counter"] -= 1
        SilaCall.clear_client_cash(force=False)


class SilaCall(BaseCall):
    """Simple resource implementation for making SiLA calls.

    Uses the sila2 python package.
    """

    __client_cash__: dict[str, SilaClient] = {}
    __context_counter__ = {"counter": 0}

    @classmethod
    def get_client(cls, address: str, port: int, insecure: bool):
        key = f"{address}:{port}"
        if key in cls.__client_cash__:
            return cls.__client_cash__[key]
        client = SilaClient(address=address, port=port, insecure=insecure)
        if cls.__context_counter__["counter"] > 0:  # Only add if in-context
            cls.__client_cash__[key] = client
        return client

    @classmethod
    def clear_client_cash(cls, force=False):
        if cls.__context_counter__["counter"] > 0 and not force:
            return  # Do nothing
        client_list = list(cls.__client_cash__.values())
        cls.__client_cash__.clear()

        for c in client_list:
            c._channel.close()
        logger.debug("Client cash cleared")

    async def Call(self, *args, **kwargs):
        if self.simulated:  # Simulation active!!
            logger.info("Simulating call to: %s", self.method)
            return

        feat, cmd = self.method.split(".")

        def get_client_and_call():
            c = SilaCall.get_client(
                address=self.url.split(":")[0],
                port=int(self.url.split(":")[1]),
                insecure=True,
            )
            call = getattr(getattr(c, feat), cmd)
            if isinstance(call, (COC, CUC)):
                h = call(*args, **kwargs)
                if not isinstance(h, (COCI, COCIWIR)):
                    return h
                try:
                    while not h.done:
                        nonasync_sleep(0.2)
                    return h.get_responses()
                finally:
                    h.cancel_execution_info_subscription()
            elif isinstance(call, (COP, CUP)):
                return call.get()
            else:
                raise Exception("This should not happened")

        try:
            return await run_non_blocking(get_client_and_call)
        except DefinedExecutionError as e:
            # The error raised comes from a SiLA server - reformat it
            raise Exception(f"{self.alias}: {e.identifier}: {e.message}")
        except UndefinedExecutionError as e:
            # The error raised comes from a SiLA server - reformat it
            raise Exception(f"{self.alias}: {e.message}")
        except Exception as e:
            # Else pass it on
            raise Exception(f"{self.alias}: {str(e)}")

    async def Ping(self) -> bool:
        try:
            return await GRPC.PingSila(f"{self.url}", timeout=1)
        except Exception:
            return False



import logging
import typing
from abc import ABCMeta, abstractmethod

from unitelabs.cdk import sila

from .types.sila_types import HandoverPosition, InvalidCommandSequence, PositionIndex

logger = logging.getLogger(__name__)


class LabwareTransferSiteControllerBase(sila.Feature, metaclass=ABCMeta):
    """
    This feature (together with the "Labware Transfer Manipulator Controller" feature) provides commands to trigger the
    sub-tasks of handing over a labware item, e.g. a microtiter plate or a tube, from one device to another in a
    standardized and generic way.

    For each labware transfer a defined sequence of commands has to be called on both involved devices to ensure the
    proper synchronization of all necessary transfer actions without unwanted physical interferences and to optimize
    the transfer performance regarding the execution time. Using the generic commands, labware transfers between any
    arbitrary labware handling devices can be controlled (a robot device has not necessarily to be involved).

    Generally, a labware transfer is executed between a source and a destination device, where one of them is the
    active device (executing the handover actions) and the other one is the passive device.

    The "Labware Transfer Site Controller" feature is used to control the labware transfer on the side of the
    passive device to hand over labware to or take over labware from an active device, which provides the
    "Labware Transfer Manipulator Controller" feature.

    If a device is capable to act either as the active or as the passive device of a labware transfer it must provide
    both features.

    The complete sequence of issued transfer commands on both devices is as follows:

    1. Prior to the actual labware transfer a "Prepare For Output" command is sent to the source device to execute all
       necessary actions to be ready to release a labware item (e.g. open a tray) and simultaneously a "Prepare For
       Input" command is sent to the destination device to execute all necessary actions to be ready to receive a
       labware item (e.g. position the robotic arm near the tray of the source device).
    2. When both devices have successfully finished their "Prepare For ..." command execution, the next commands are
       issued.
    3a If the source device is the active device it will receive a "Put Labware" command to execute all necessary
       actions to put the labware item into the destination device. After the transfer has been finished successfully,
       the destination device receives a "Labware Delivered" command, that triggers all actions to be done after the
       labware item has been transferred (e.g. close the opened tray).
    3b If the destination device is the active device it will receive a "Get Labware" command to execute all necessary
       actions to get the labware from the source device (e.g. gripping the labware item). After that command has been
       finished successfully, the source device receives a "Labware Removed" command, that triggers all actions to be
       done after the labware item has been transferred (e.g. close the opened tray).

    The command sequences for a passive source or destination device have always to be as follows:
    - for a passive source device:        PrepareForOutput - LabwareRemoved
    - for a passive destination device:   PrepareForInput - LabwareDelivered

    If the commands issued by the client differ from the respective command sequences an "Invalid Command Sequence"
    error will be raised.

    To address the location, where a labware item can be handed over to or from other devices, every device must
    provide one or more uniquely named positions (handover positions) via the "Available Handover Positions" property.
    A robot (active device) should have at least one handover position for each device that it interacts with, whereas
    most passive devices will only have one handover position. In the case of a position array (e.g. a rack), the
    position within the array is specified via the sub-position of the handover position, passed as an index number.

    To address the positions within a device where the transferred labware item has to be stored at or is to be taken
    from (e.g. the storage positions inside an incubator), the internal position is specified. Each device must provide
    the number of available internal positions via the "Number Of Internal Positions" property. In the case of no
    multiple internal positions, this property as well as the "Internal Position" parameter value must be 1.

    With the "Prepare For Input" command there is also information about the labware transferred, like labware type or
    a unique labware identifier (e.g. a barcode).
    """

    @abstractmethod
    @sila.ObservableCommand(name="Prepare For Input", errors=[InvalidCommandSequence])
    async def PrepareForInput(
        self,
        HandoverPosition: HandoverPosition,
        InternalPosition: PositionIndex,
        LabwareType: str,
        LabwareUniqueID: str,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to accept new labware at the specified handover position.

        .. parameter:: Indicates the position where the labware will be handed over.
        .. parameter:: Indicates the position which the labware will be stored at within the device, e.g. internal
                       storage positions of an incubator.
        .. parameter:: Specifies the type of labware that will be handed over to transfer information about the
                       labware that the device might need to handle it correctly.
        .. parameter:: Represents the unique identification of a labware in the controlling system. It is assigned
                       by the system and must remain unchanged during the whole process.
        """

    @sila.ObservableCommand(name="Prepare For Output", errors=[InvalidCommandSequence])
    async def PrepareForOutput(
        self,
        HandoverPosition: HandoverPosition,
        InternalPosition: PositionIndex,
        *,
        status: sila.Status,
    ) -> None:
        """
        Put the device into a state in which it is ready to release the labware at the specified handover position.

        .. parameter:: Indicates the position where the labware will be handed over.
        .. parameter:: Indicates the position which the labware will be retrieved from within the device, e.g. internal \
              storage positions of an incubator.

        """

    @abstractmethod
    @sila.ObservableCommand(name="Labware Delivered", errors=[InvalidCommandSequence])
    async def LabwareDelivered(
        self,
        HandoverPosition: HandoverPosition,
        *,
        status: sila.Status,
    ):
        """
        Notifies the passive destination device of a labware item that has been transferred to it (sent after a "Prepare \
              For Input" command)

        .. parameter:: Indicates the position the labware item has been delivered to.
        """

    @abstractmethod
    @sila.ObservableCommand(name="Labware Removed", errors=[InvalidCommandSequence])
    async def LabwareRemoved(
        self,
        HandoverPosition: HandoverPosition,
        *,
        status: sila.Status,
    ):
        """
        Notifies the passive source device of a labware item that has been removed from it (sent after a "Prepare For \
              Output" command).

        .. parameter:: Indicates the position the labware has been removed from.
        """

    @abstractmethod
    @sila.UnobservableProperty(name="Available Handover Positions")
    async def AvailableHandoverPositions(self) -> typing.List[HandoverPosition]:
        """
        All handover positions of the device including the number of sub-positions.
        """

    @sila.UnobservableProperty(name="Number Of Internal Positions")
    async def NumberOfInternalPositions(
        self,
    ) -> typing.Annotated[int, sila.constraints.MinimalInclusive(1)]:
        """
        The number of internal positions the device has.
        """
        return 1  # Default (not used)



import abc

try:
    from unitelabs.cdk import sila
except ImportError as ex:
    raise ImportError(
        "Please install the unitelabs package by running 'pip install sila2-feature-lib[unitelabs]'"
    ) from ex


class InvalidParameterError(Exception):
    """The given parameter is invalid."""


class ReportGenerationError(Exception):
    """An error occurred during report generation."""


class InternalError(Exception):
    """An internal error occurred."""


class ReportGenController(sila.Feature, metaclass=abc.ABCMeta):
    def __init__(
        self,
        *args,
        identifier="ReportGenController",
        display_name="Report Generator Controller",
        description="Report generation SiLA feature",
        **kwargs,
    ):
        super().__init__(
            *args,
            identifier=identifier,
            display_name=display_name,
            description=description,
            **kwargs,
        )

    @abc.abstractmethod
    @sila.UnobservableCommand(
        name="Generate Report",
        errors=[InternalError, InvalidParameterError, ReportGenerationError],
    )
    @sila.Response("Report ID")
    async def generate_report(
        self,
        identifier: str,
        additional_info: str,
    ) -> str:
        """
        Generate a report from an identifier.

        .. parameter:: Unique indentifier of a data set to generated the report from
        .. parameter:: Additional information needed to generate the report

        .. return:: Unique identifier of the generated report
        """
        # Note: Abstract method



import json
import logging
from importlib import import_module
from typing import Generic, TypeVar, Union

try:
    from unitelabs.cdk import sila
except ImportError as ex:
    raise ImportError(
        "Please install the unitelabs package by running 'pip install sila2-feature-lib[unitelabs]'"
    ) from ex

logger = logging.getLogger(__name__)

VERSION = "0.1"


class ConnectionHandleBase:
    """Base class for defining communication methods."""

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Close the connection and release any resources."""
        pass

    def get_connection_details(self):
        """Return a dictionary with connection details."""
        raise NotImplementedError

    def open(self):
        """Open the connection."""
        return self

    async def test(self):
        """Test the connection."""
        raise NotImplementedError


T = TypeVar("T")


class DataStoreService(sila.Feature, Generic[T]):
    """Feature for accessing shared resources"""

    # Static content
    content = {"handle": None}

    def __init__(
        self,
        identifier="DataStoreService",
        display_name="Data Store Service",
        description="Feature for accessing shared resources",
        version=VERSION,
        handle: Union[str, T] = None,
        **kwargs,
    ):
        super().__init__(
            identifier=identifier,
            display_name=display_name,
            description=description,
            version=version,
            **kwargs,
        )
        if handle:
            self.assign_handle(handle)

    def assign_handle(self, handle: Union[str, T]):
        if isinstance(handle, str):
            m = import_module(".".join(handle.split(".")[:-1]))
            handle = getattr(m, handle.split(".")[-1])()

        if not isinstance(handle, ConnectionHandleBase):
            raise ValueError(f"Invalid handle type {type(handle)}")

        DataStoreService.content["handle"] = handle
        return self

    @classmethod
    def get_handle(cls) -> T:
        """Get a handle to access shared resources."""
        return DataStoreService.content["handle"]

    @sila.UnobservableCommand(
        identifier="TestConnection",
        name="Test Connection",
    )
    @sila.Response("Status", "Status of the connection")
    async def test_connection(self) -> str:
        # Just open the connection and the close it
        with DataStoreService[ConnectionHandleBase].get_handle() as conn:
            await conn.test()  # Raise on error
        return "OK"

    @sila.UnobservableProperty(
        identifier="Details",
    )
    async def details(self) -> str:
        return json.dumps(
            DataStoreService[ConnectionHandleBase].get_handle().get_connection_details()
        )



from os import environ
from typing import Any

import httpx

from ..feature_ul import ConnectionHandleBase

UUID = str


class ResourceHandle(ConnectionHandleBase):
    """Default implementation of a connection handle to a Rest API style data service."""

    def __init__(self, timeout=10, url: str = None, **kwargs):
        self.timeout = timeout
        self.kwargs = kwargs
        # If no URL is provided, try to get it from the environment
        # Example format: "http://localhost:8000"
        self.url = url or environ.get("APP_DATASTORE_URL", None)

    async def custom_call(self, method="GET", endpoint: str = "/", **kwargs):
        """Make a custom call, anything not covered below."""
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            rpl = await client.request(method, url=self.url + endpoint, **kwargs)
            return rpl.raise_for_status()

    def get_connection_details(self):
        """Return connection details."""
        return {"url": self.url}

    async def get_info(self, uuid: UUID) -> dict[str, Any]:
        """Get information about a resource by its UUID."""
        async with httpx.AsyncClient(timeout=self.timeout, **self.kwargs) as client:
            rpl = await client.get(self.url + "/get_info", params={"uuid": uuid})
            return rpl.raise_for_status().json()

    async def get_resources(
        self,
        resource_type: str,
        filter: dict[str, Any] = None,
    ) -> list[UUID]:
        """Returns a list of UUIDs of all matching resources."""
        async with httpx.AsyncClient(timeout=self.timeout, **self.kwargs) as client:
            rpl = await client.post(
                self.url + "/get_resources",
                params={"resource_type": resource_type},
                json={} if filter is None else filter,
            )
            return rpl.raise_for_status().json()

    async def test(self) -> None:
        """Ping the server to check if it is alive/available."""
        async with httpx.AsyncClient(timeout=self.timeout, **self.kwargs) as client:
            rpl = await client.get(self.url + "/test")
            rpl.raise_for_status()

    async def update_resource(
        self,
        uuid: UUID,
        changes: dict[str, Any],
    ) -> None:
        """Updated properties of a resource by its UUID."""
        async with httpx.AsyncClient(timeout=self.timeout, **self.kwargs) as client:
            rpl = await client.post(
                self.url + "/update_resource",
                params={"uuid": uuid},
                json=changes,
            )
            rpl.raise_for_status()



import unittest


class TestXMLImport(unittest.TestCase):
    def test_xml_load(self):
        from sila2_feature_lib import get_xml, get_xml_path

        _ = get_xml, get_xml_path  # Avoid unused import warning


class TestImport(unittest.TestCase):
    def test_datastore(self):
        from sila2_feature_lib.datastore.v001_0.feature_ul import DataStoreService

        _ = DataStoreService

    def test_labwaremanipulator(self):
        from sila2_feature_lib.labwaremanipulator.v001_0.feature_ul import (
            LabwareTransferManipulatorControllerBase,
        )

        _ = LabwareTransferManipulatorControllerBase

    def test_labwaresite(self):
        from sila2_feature_lib.labwaresite.v001_0.feature_ul import (
            LabwareTransferSiteControllerBase,
        )

        _ = LabwareTransferSiteControllerBase

    def test_reportgen(self):
        from sila2_feature_lib.reportgen.v001_0.feature_ul import ReportGenController

        _ = ReportGenController

    def test_resource(self):
        from sila2_feature_lib.resource.v001_0.feature_ul import ResourcesService

        _ = ResourcesService

    def test_simulation(self):
        from sila2_feature_lib.simulation.v001_0.feature_ul import SimulatorController

        _ = SimulatorController

    def test_workflowrunner(self):
        from sila2_feature_lib.workflowrunner.v001_0.feature_ul import (
            get_workflow_feature,
        )

        _ = get_workflow_feature()
